//
//  Copyright (C) 1997 by S.Tomabechi
//  All rights reserved.
//
//  class for Mudular Multiplication Without Trial Division by Montgomery
//
//  not faster than usual modular multiplication with trial division (why ?)
//
#ifndef __MONT_HPP
#define __MONT_HPP
#define _ONKY_MINTDECL_
#include "Mint.hpp"
#include "Error.hpp"

class Mont
{
private:
  Mint   rep ;              // representative  xR mod N for integer x
public:
  static bool SetModulus(const Mint& N) ;
         operator  Mint() const ; // conversion operator
  Mont(const Mint& x) ;           // Mont <- reduce(Mint)
//  Mont(const Mont& x) ;           // copy constructor
  Mont(const long  n) ;
  Mont::Mont(const Mont& x):rep(x.rep){}
  Mont() {}
  ~Mont(){}
  Mont& operator/=(const Mont& x) ;
  Mont& operator/(const Mont& x) ;

  Mont& operator=(const Mint& x) ;
  Mont& operator=(const Mont& x) ;
  Mont& operator=(const long n) ;
  Mont  operator*(const Mont& x) ;
  Mont& operator*=(const Mont& x) ;
  Mont  operator+(const Mont& x) ;
  Mont& operator+=(const Mont& x) ;
  Mont  operator-(const Mont& x) ;
  Mont& operator-=(const Mont& x) ;
  Mont& power(const Mont& x,const Mint& y) ;
protected:
  static long mantissa ;
  static Mint modulus ;     // N
  static Mint inverse ;     // N^-1
  static Mint IR ;          // R^-1
  static Mont _Motemp ;
  Mont& reduce(const Mint& x) ;   // Montgomery reduction
  Mont& reduce2(const Mint& x) ;   // Montgomery reduction
  Mont& isAdd(const Mont& x,const Mont& y) ;
  Mont& isSub(const Mont& x,const Mont& y) ;
  Mont& isMul(const Mont& x,const Mont& y) ;
  Mont& isDiv(const Mont& x,const Mont& y) ;
  Mont& isInverse(const Mont& x) ;

  friend Mont power(const Mont& x,const Mint& n) ;
} ;

inline Mont::Mont(const long n)
{
  *this=n ;
}

inline Mont& Mont::isAdd(const Mont& x,const Mont& y)
{
  rep.isAdd(x.rep,y.rep) ;
  return rep.nDigit>=mantissa ? (rep.isSub(rep,modulus),*this):*this ;
//  if( rep.nDigit>mantissa ) rep.isSub(rep,modulus) ;
//  if( rep>=modulus ) rep.isSub(rep,modulus) ;
//  return *this ;
}

inline Mont& Mont::isSub(const Mont& x,const Mont& y)
{
  rep.isSub(x.rep,y.rep) ;
  return rep.isNegative() ? (rep.isAdd(rep,modulus),*this):*this ;
//  if( rep.isNegative() ) rep.isAdd(rep,modulus) ;
//  rep.isNegative() ? rep.isAdd(rep,modulus):rep ;
//  return *this ;
}


inline Mont Mont::operator*(const Mont& a)
{
  return _Motemp.isMul(*this,a) ;
}
inline Mont& Mont::operator*=(const Mont& a)
{
  return isMul(*this,a) ;
}

inline Mont& Mont::operator/=(const Mont& x)
{
  return isDiv(*this,x) ;
}

inline Mont& Mont::operator/(const Mont& x)
{
  return _Motemp.isDiv(*this,x) ;
}

inline Mont& Mont::operator+=(const Mont& x)
{
  return isAdd(*this,x) ;
}

inline Mont Mont::operator+(const Mont& x)
{
  return _Motemp.isAdd(*this,x) ;
}

inline Mont& Mont::operator-=(const Mont& x)
{
  return isSub(*this,x) ;
}

inline Mont Mont::operator-(const Mont& x)
{
  return _Motemp.isSub(*this,x) ;
}

inline Mont::operator Mint() const
{
  return (_Motemp.reduce(rep)).rep ;
}


Mont power(const Mont& x,const Mint& y) ;
inline Mont power(const Mont& x,const Mint& n)
{
  return Mont::_Motemp.power(x,n) ;
}

#if 0
inline Mont::Mont(const Mont& x)
{
  rep=x.rep ;

}
#endif

inline Mont& Mont::isMul(const Mont& x,const Mont& y)
{
  return reduce(Mint::_Mtemp.isMul(x.rep,y.rep)) ;
}

inline Mont::Mont(const Mint& x)
{
  rep.isRem(Mint::_Mtemp.isShift32(x,mantissa),modulus,PositiveRem) ;
}

inline Mont& Mont::operator=(const long n)
{
  rep.isRem((rep=n)[mantissa],modulus,PositiveRem) ;
  return *this ;
}

inline Mont& Mont::operator=(const Mint& x)
{
  rep.isRem(Mint::_Mtemp.isShift32(x,mantissa),modulus,PositiveRem) ;
  return *this ;
}

inline Mont& Mont::operator=(const Mont& x)
{
  rep=x.rep ;
  return *this ;
}

inline Mont& Mont::isDiv(const Mont& x,const Mont& y)
{
  _Motemp.rep.isMul(x.rep,Mint::_Mtemp.isInvModp((_Motemp.reduce(y.rep)).rep,modulus)) ;
  rep.isRem(_Motemp.rep,modulus,PositiveRem) ;
  return *this ;
}

inline Mont& Mont::isInverse(const Mont& x)
{
  Mint::_Mtemp.isInvModp((_Motemp.reduce(x.rep)).rep,modulus) ;
  rep.isRem(Mint::_Mtemp,modulus,PositiveRem) ;
  return *this ;
}

inline Mont&  Mont::reduce2(const Mint& x)
{
#if 1
    rep.isMul(x,inverse) ;
    _Motemp.rep.isRem2(rep,(mantissa)<<5) ;
//cout << endl << rep ;
    rep.isMul(_Motemp.rep,modulus) ;
    rep.isAdd(rep,x) ;
//  rep.isRem2(rep,(mantissa)<<5) ;
//    rep.isShift32(rep,-mantissa) ;
    rep[-mantissa] ;
    return rep>=modulus ? (rep.isSub(rep,modulus),*this):*this ;
#else
    rep.isMul(x,IR) ;
    rep.isRem(rep,modulus) ;
    return *this ;
//  return rep>=modulus ? (rep.isSub(rep,modulus),*this):*this ;
#endif
}
#endif