//
//	Copyright (C) 1996 by S.Tomabechi
//	All rights reseved.
//
//	calculate \pi by Chudnovsky's formula
//
//	 426880*sqrt(10005) 			 (6n)!(545140134n+13591409)
//	 \\\\\\\\\  \sigma	\\\\\\\\\\\\\\\
//		  \pi				n>=0	 (n!)^3*(3n)!*(-640320)^(3n)
//
#include "Mint.hpp"
#include "Fint.hpp"
#include "FintFunc.hpp"

Fint Chudnovsky(int precision)
{
  static int fdigit=-1 ;
  precision=GetPrecision()>precision ? GetPrecision():precision ;

  // already calculated pi with enough precision
  if( fdigit>=precision ) return Fint::_pi(precision-fdigit) ;

  Fint p(precision+2,precision+1) ;
  unsigned long m ;
  long c(-320160L) ;

  unsigned long j ;
  Mint n6(1) ;
  Mint n3(1) ;
  Mint n1(1) ;
  Mint cn(1) ;
  Mint c3(1) ;
  Mint n(1) ;
  Mint c2("6541681608") ;

  Mint a(13591409UL) ;
  unsigned long b(545140134UL) ;
  Fint d(precision+2,precision+1) ;
  d=1 ;
  ((c3*=c)*=c)*=c ; // c3=c^3
  p=a ;
  for( m=0,j=1 ;; j++) {
	a+=b ;
//	  m=6*j ;
	m+=6 ;
	d*=(m-1) ;		  // Aq (6n)! ,(3n)! ̂
	d*=(m-3) ;		  // ʂ̈q͏
	d*=(m-5) ;
	d/=(c3*j*j*j) ;
	if( d==0 ) break ;
	p+=a*d ;
  }
  p/=b ;

  Fint::_pi.fDigit=precision ;
  Fint::_pi=-8*c3 ;

  Fint::_pi.isSqrt(Fint::_pi)/=(p*c2) ;

  fdigit=precision ;					// ݂̐xoĂ
#ifdef _MEMMAN_DEBUG
Fint::_pi.check(__FILE__,__LINE__) ;
#endif
  return Fint::_pi[precision-Fint::_pi.fdigit()] ;	// round extra digits

}

