//
char _Copyright[]="Copyright (C) by S.Tomabechi 2000" ;
//      All rights reseved.
//
//  Ellicptic Curve Primality Proving
//
//  c.f. A.O.L. Atkin and F.Morain.
//       Elliptic curves and primality proving
//       math. Comp.,61,29-67,1993
//

#include <algorith.h>
#include "Mint.hpp"
#include "Fint.hpp"
#include <vector.h>
#include <time.h>
#include "PolyM.hpp"
#include "MintFunc.hpp"
#include "Error.hpp"
#include "CPUTime.hpp"

Mint aa,bb;
Mint b ;
Mint a ;
Mint Nj ;
using namespace std ;

Mint parse(char* exp) ;
unsigned long hDsum1(unsigned long D,double& sum);
int Pocklington(Mint& N);
int Lucas(Mint& N);

void cal(Mint& R,unsigned long D,Mint& J);
int RootModp(PolyM& f,Mint& N,vector< Mint >& roots);
int RootModp1(PolyM& f,Mint& N,vector< Mint >& roots);
int RootModp(PolyFp& f,vector< Fp >& roots) ;
PolyM gcd(PolyM a,PolyM b,Mint& N);
bool Find_a_root(PolyM& T1,Mint& N,Mint& R);
bool isNegDisc(unsigned long D);
bool Cornacchia(const Mint& d,const Mint& p,Mint& x,Mint& y);
void HilbertPolynomial(unsigned long D,PolyM& Pm);
void WP_new(unsigned long D,PolyM& Pm);
void Weber(const unsigned long D,PolyM& Pm);
void JmodN(Mint& R,Mint& N,unsigned long D,Mint& j);
bool Cornacchia4(Mint& d,Mint& p,Mint& x,Mint& y);
PolyM power(const PolyM& x,const Mint& n,const PolyM& p,const Mint& N);


//
// # of automorphism
//
unsigned long wD(unsigned long D)
{
  switch( D ) {
    case 3:
      return 6 ;
    case 4:
      return 4 ;
    default:
      return 2 ;
  }
}

// (x,y)=(x,y)+(wx,wy)
bool add(
         Mint& N,
         Mint& x,Mint& y,
         Mint& wx,Mint& wy,
         Mint& d)
{
  Mint w,lmd ;
  Mint ww ;
//  w.isInvModp(wx-x,N) ;
#if 0
if( (y*y-x*x*x-a*x-b)%N!=0 ) Error(a555) ;
if( (wy*wy-wx*wx*wx-a*wx-b)%N!=0 ) Error(a555) ;
#endif
  w=invmodp(wx-x,N) ;
  if( w==0 ) {
    d=gcd(wx-x,N) ;
    return false ;
  }
  lmd=mod((wy-y)*w,N) ;
  w=x ;
  x=mod(lmd*lmd-x-wx,N) ;
  y=mod(lmd*(w-x)-y,N) ;
#if 0
if( (y*y-x*x*x-a*x-b)%N!=0 ) Error(a7a7a) ;
#endif
  return true ;
}

// (x,y)=2*(x,y)
bool dbl(Mint& a,Mint& N,
         Mint& x,Mint& y,
         Mint& d)
{
  Mint w,lmd ;
  Mint ww ;
#if 0
if( (y*y-x*x*x-a*x-b)%N!=0 ) Error(a555) ;
#endif
  w=invmodp(2*y,N) ;
  if( w==0 ) {
    d=gcd(y,N) ;
    return false ;
  }
  lmd=mod((3*x*x+a)*w,N) ;
  w=x ;
  x=mod(lmd*lmd-2*x,N) ;
  y=mod(lmd*(w-x)-y,N) ;
#if 0
if( (y*y-x*x*x-a*x-b)%Nj!=0 ) Error(a7a7a) ;
#endif
  return true ;                  // normal end
}

// (x,y)=m*(x,y)
bool mult(Mint& a,Mint& N,Mint m,Mint& x,Mint& y,Mint& d)
{

//cout << endl << "mult " << x << " " << y;
#if 0
  if( (y*y-x*x*x-a*x-b)%N!=0 ) Error(a7a7a) ;
#endif  
  Mint wx,wy,q ;
  while( (long(m)&1)==0 ) {  // m is even
    m>>=1 ;
    if( !dbl(a,N,x,y,d) ) return false ;
  }
  wx=x ;
  wy=y ;
  m>>=1 ;
  while( m>0 ) {
    if( !dbl(a,N,wx,wy,d) ) {
        return false ;
    }
    if( long(m)&1 ) {         // m is odd
      if( !add(N,x,y,wx,wy,d) ) {
        if( m==1 ) { x=0 ; }
        return false ;
      }
    }
    m>>=1 ;
  }
#if 0
if( (y*y-x*x*x-a*x-b)%N!=0 ) Error(a7a7a) ;
#endif
  return true ;                  // normal end
}


unsigned long nextDisc(unsigned long D)
{
  for( ;; D++ ) {
    if( isNegDisc(D) ) return D ;
  }
}

int divisor(const Mint& a,Mint& t,vector<Mint>& d)
{
  Mint b ;
  int itry ;
  unsigned long p ;
  vector<Mint> s ;
  bool found=false ;

  if( EulerCriterion2(a) /* && APRCL(a) */ )   return -1 ;  // a is prime

//  while( !d.empty() )         // stack ɂ
//    d.pop_back() ;
  s.push_back(a) ;            // 𐮐X^bNɐς

  for( ; !s.empty() ; ) {
    if( s.back()==1 ) {       // 1 Ȃ 𐮐X^bN珜
      s.pop_back() ;
      continue ;
    }
    p=prmdiv(s.back(),SMALL_DIVISOR) ;      // fT
    if( p!=0 )  {
      if( p==1 ) {            // f
        d.push_back(s.back()) ;
        s.pop_back() ;
      }
      else {                  // f݂
        s.back()/=p ;
        d.push_back(Mint(p)) ;// fX^bN
        if( t<b ) goto RET0 ;
      }
    }
    else break ;              // ͂⏬f͂Ȃ
  }


  for( ; !s.empty() ; ) {
//break ;
    if( s.back()==1 ) {
      s.pop_back() ;
      continue ;
    }


//    if( ilog2(s.back())<202 ) itry=1 ;
//    else itry=2 ;
     itry=100 ;

    if( EulerCriterion2(s.back())  /* && APRCL(s.back()) */ ) {  // cĂ鐔́Af ?
      d.push_back(s.back()) ;  // fX^bNɈړ
      s.pop_back() ;
      if( d.back()>t ) found=true ;
      if( found ) goto RET0 ;
      continue ;               // ܂cĂ鐔𒲂ׂ
    }
    b=rho(s.back(),itry,1) ;   // fłȂ̂ŕ݂ ϖ@
    if( b==1 ) {               // łȂ
break ;
      b=rho(s.back(),itry,-1) ;// p^ւĎ
      if( b==1 ) b=rho(s.back(),itry,3) ;
    }
    if( b!=1 ) {               // ݂
      if( EulerCriterion2(b) ) {   // ͑f ?
        d.push_back(b) ;
        s.back()/=b ;
        if( t<b ) found=true ;
      }
      else {                   // ͑fłȂ
        s.back()/=b ;
        s.push_back(b) ;       // 𐮐X^bNɐς
      }
    }
    else break ;
  }

  for( ; !s.empty() ; ) {
break ;
    if( s.back()==1 ) {
      s.pop_back() ;
      continue ;
    }

    if( EulerCriterion2(s.back()) /*  && APRCL(s.back())*/ ) {  // cĂ鐔́Af ?
      d.push_back(s.back()) ;
      s.pop_back() ;
      if( d.back()>t ) found=true ;
      if( found ) goto RET0 ;
      continue ;            // ܂cĂ鐔𒲂ׂ
    }
    b=p_1(s.back(),250000) ; // fłȂ̂ŕ݂ p-1@
    if( b!=1 ) {
      if( EulerCriterion2(b) ) {
        d.push_back(b) ;
        s.back()/=b ;
        if( t<b ) found=true ;
      }
      else {
        s.back()/=b ;
        s.push_back(b) ;
      }
    }
    else break ;
  }



  for( ; !s.empty() ; ) {
break ;
    if( s.back()==1 ) {
      s.pop_back() ;
      continue ;
    }

    itry=(s.back()).ndigit()-3 ;
    if( itry<=1 ) itry=2 ;
itry=2 ;

    if( EulerCriterion2(s.back())  && APRCL(s.back()) ) {  // cĂ鐔́Af ?
      d.push_back(s.back()) ;
      s.pop_back() ;
      if( d.back()>t ) found=true ;
      if( found ) goto RET0 ;
      continue ;            // ܂cĂ鐔𒲂ׂ
    }
    b=ecm(s.back(),itry) ;  // fłȂ̂ŕ݂ ȉ~Ȑ@
    if( b!=1 ) {            // Ȃ邱Ƃ͂Ă͂ȂȂ
                            // ֗̂́AԂ񎸔sꍇ
      if( EulerCriterion2(b) ) {
        d.push_back(b) ;
        s.back()/=b ;
        if( t<b ) found=true ;
      }
      else {
        s.back()/=b ;
        s.push_back(b) ;
      }

    }
    else break ;
  }

#if 0
  for(; !s.empty() ;) {
break ;
    if( s.back()==1 ) {
      s.pop_back() ;
      continue ;
    }

    if( EulerCriterion2(s.back()) && APRCL(s.back())  ) {  // cĂ鐔́Af ?
      d.push_back(s.back()) ;
      s.pop_back() ;
      continue ;            // ܂cĂ鐔𒲂ׂ
    }
    b=PPMPQS(s.back()) ;    // fłȂ̂ŕ  񎟂ӂ邢@
    if( b==1 ) {            // Ȃ邱Ƃ͂Ă͂ȂȂ
                            // ֗̂́AԂ񎸔sꍇ
      if( EulerCriterion2(s.back()) /* &&  APRCL(s.back()) */ ) {
        d.push_back(b) ;
        s.back()/=b ;
        continue ;
      }
      else return -1 ;
    }
    else {                  // ݂
      if( EulerCriterion2(b) && APRCL(b) ) {
        d.push_back(b) ;   // fȂfzɓ
        s.back()/=b ;
      }
      else {
        s.back()/=b ;      // fłȂ΃X^bNɐς
        s.push_back(b) ;
      }
    }
  }
#endif
RET0:

                    // ꂽfɕבւ
  sort(d.begin(),d.end()) ;

                            // Z
//  b=1 ;
//  for( j=0 ; j<d.size() ; j++ ) b*=d[j] ;
//  if( b!=a ) Error( division error);
//  SetFormat(form) ;

  return d.size() ;    // f̌

}


int divisor1(
  const Mint& a,Mint& t,unsigned short* ptbl,unsigned long* rtbl,
  unsigned long tblMax,Mint& x,
   vector<Mint>& d)
{
  Mint b ;
  int itry ;
  unsigned long p ;
  vector<Mint> s ;
  bool found=false ;
  unsigned long k ;

  if( EulerCriterion2(a) /* && APRCL(a) */ )   return -1 ;  // a is prime

//  while( !d.empty() )         // stack ɂ
//    d.pop_back() ;
  s.push_back(a) ;            // 𐮐X^bNɐς

  for( ; !s.empty() ; ) {
    if( s.back()==1 ) {       // 1 Ȃ 𐮐X^bN珜
      s.pop_back() ;
      continue ;
    }
    p=0 ;
    for( k=0 ; k<tblMax ; k++  ) {
//      p=nextprm(p) ;
      p+=ptbl[k] ;
//cout << " k " << k ;
      if( mod(x,p)==rtbl[k] ) {
//cout << " " << p ;
        s.back()/=p ;

        if( Mint::_res!=0 ) {
cout << endl << "a " << a ;
cout << endl << "x " << x ;

            Error(div error) ;
        }
        d.push_back(Mint(p)) ;// fX^bN
        while( s.back()%p==0 ) s.back()/=p ;
        if( t<b ) goto RET0 ;
      }
    }
    break ;
  }
  for( ; !s.empty() ; ) {
//break ;
    if( s.back()==1 ) {
      s.pop_back() ;
      continue ;
    }


//    if( ilog2(s.back())<202 ) itry=1 ;
//    else itry=2 ;
     itry=1000 ;
     itry=100 ;

    if( EulerCriterion2(s.back())  /* && APRCL(s.back()) */ ) {  // cĂ鐔́Af ?
      d.push_back(s.back()) ;  // fX^bNɈړ
      s.pop_back() ;
      continue ;               // ܂cĂ鐔𒲂ׂ
    }
    b=rho(s.back(),itry,1) ;   // fłȂ̂ŕ݂ ϖ@
    if( b==1 ) {               // łȂ
break ;
      b=rho(s.back(),itry,-1) ;// p^ւĎ
      if( b==1 ) b=rho(s.back(),itry,3) ;
    }
    if( b!=1 ) {               // ݂
      if( EulerCriterion2(b) ) {   // ͑f ?
        d.push_back(b) ;
        s.back()/=b ;
      }
      else {                   // ͑fłȂ
        s.back()/=b ;
        s.push_back(b) ;       // 𐮐X^bNɐς
      }
    }
    else break ;
  }


//  if( EulerCriterion2(s.back())  ) {  // cĂ鐔́Af ?
//    d.push_back(s.back()) ;
//    s.pop_back() ;
//  }



RET0:

                    // ꂽfɕבւ
  sort(d.begin(),d.end()) ;

                            // Z
//  b=1 ;
//  for( j=0 ; j<d.size() ; j++ ) b*=d[j] ;
//  if( b!=a ) Error( division error);
//  SetFormat(form) ;

  return d.size() ;    // f̌

}



// N-1 ̂Ƃ res= 2 f
// N ̂Ƃ res= 0 f
int divisor2(
  const Mint& N,unsigned short* ptbl,unsigned long* rtbl,
  unsigned long tblMax,
  const unsigned long res ,
   vector<Mint>& d)
{
  Mint b ;
  int itry ;
  unsigned long p ;
  vector<Mint> s ;
  bool found=false ;
  unsigned long k ;

  if( EulerCriterion2(N) /* && APRCL(a) */ )   return -1 ;  // a is prime

  s.push_back(N) ;            // 𐮐X^bNɐς

  for( ; !s.empty() ; ) {
    if( s.back()==1 ) {       // 1 Ȃ 𐮐X^bN珜
      s.pop_back() ;
      continue ;
    }
    p=0 ;
    p+=ptbl[0] ;
    if( (rtbl[0]-res)%2==0 ) {
      s.back()/=p ;

      if( Mint::_res!=0 ) {

          Error(div error) ;
     }

      d.push_back(Mint(p)) ;// fX^bN
      while( s.back()%p==0 ) s.back()/=p ;
    }

    for( k=1 ; k<tblMax ; k++  ) {
//      p=nextprm(p) ;
      p+=ptbl[k] ;
      if( rtbl[k]!=res ) continue ;
      s.back()/=p ;

      if( Mint::_res!=0 ) {

          Error(div error) ;
     }

      d.push_back(Mint(p)) ;// fX^bN
      while( s.back()%p==0 ) s.back()/=p ;
    }
    break ;
  }


  for( ; !s.empty() ; ) {
//break ;
    if( s.back()==1 ) {
      s.pop_back() ;
      continue ;
    }


//    if( ilog2(s.back())<202 ) itry=1 ;
//    else itry=2 ;
     itry=1000 ;
     itry=100 ;

    if( EulerCriterion2(s.back())  /* && APRCL(s.back()) */ ) {  // cĂ鐔́Af ?
      d.push_back(s.back()) ;  // fX^bNɈړ
      s.pop_back() ;
      continue ;               // ܂cĂ鐔𒲂ׂ
    }
    b=rho(s.back(),itry,1) ;   // fłȂ̂ŕ݂ ϖ@
    if( b==1 ) {               // łȂ
break ;
      b=rho(s.back(),itry,-1) ;// p^ւĎ
      if( b==1 ) b=rho(s.back(),itry,3) ;
    }
    if( b!=1 ) {               // ݂
      if( EulerCriterion2(b) ) {   // ͑f ?
        d.push_back(b) ;
        s.back()/=b ;
      }
      else {                   // ͑fłȂ
        s.back()/=b ;
        s.push_back(b) ;       // 𐮐X^bNɐς
      }
    }
    else break ;
  }




//  if( EulerCriterion2(s.back())  ) {  // cĂ鐔́Af ?
//    d.push_back(s.back()) ;
//    s.pop_back() ;
//  }



RET0:

                    // ꂽfɕבւ
  sort(d.begin(),d.end()) ;

                            // Z
//  b=1 ;
//  for( j=0 ; j<d.size() ; j++ ) b*=d[j] ;
//  if( b!=a ) Error( division error);
//  SetFormat(form) ;

  return d.size() ;    // f̌

}

//q|N-1=m
bool Pocklington1(Mint& N,Mint& m,Mint& q)
{
  unsigned long a ;
  Mint M,G ;

  for( a=2 ; a<=13000 ; a++ ) {
cout << endl << "a " << a ;
    M=a ;
    M=power(M,m/q,N) ;                    // B^S mod N
cout << "  " << power(M,q,N) ;
//    if( power(M,q,N)!=1 ) return false ;   // B^S^R mod N
    if( power(M,q,N)!=1 ) continue ;   // B^S^R mod N
    G=gcd(M-1,N) ;                      //B^S-1 mod N
    if( G==1 ) return true ;               // ̂Ƃ U ̑f N ͑f
    if( G<N ) return false ;
  }
  return false ;   // ?
}


// calculate Lucas sequence mod N
void Lucas1(const Mint& N,Mint& P,Mint& Q,Mint n,
           Mint& U,Mint& V)
{
  Mint D=P*P-4*Q ;
  Mint a1(1),a2(0),w1(P),w2(1) ;
  Mint t ;
  unsigned long m(0) ;
  unsigned long l=len(n) ;
  Mint half=invmodp(Mint(2),N) ;
  for( ; l>0 ; l-- ) {

    if( ((unsigned long)n)&1 ) {
      t=(a1*w1+D*a2*w2) ;
      a2=(a1*w2+a2*w1)^N ;
      a1=t^N ;
      m++ ;
    }

    t=(w1*w1+D*w2*w2) ;      // ̐
    (t*=half)^=N ;
    (w2*=w1)^=N ;                 //  D ̌W
    w1=t ;
    n>>=1 ;


  }

  U=(a2*power(half,Mint(m-1),N))%N  ;
  V=(a1*power(half,Mint(m-1),N))%N  ;
}




bool LucasTest1(Mint& N,Mint& m,Mint& q)
{
  Mint P,Q ;
  Mint U,V ;
  P=1 ;
  for( Q=-100002 ; Q<100000 ; Q++ ) {
    if( Kronecker(P*P-4*Q,N)!=-1 ) continue ;
//    LucasSequence(P,Q,(N+1)/2,U,V) ;
    Lucas1(N,P,Q,m,U,V) ;
//V=U ;
//cout << endl << "Lucas " << U ;
    if( U%N!=0 ) continue ;
    Lucas1(N,P,Q,q,U,V) ;
//V=U ;
//cout << " " << U << " gcd " << gcd(U,N) ;
    if( gcd(U,N)==1 ) return true ;
  }

  P=2 ;
  for( Q=-100000+1 ; Q<100000 ; Q+=2 ) {
    if( Kronecker(P*P-4*Q,N)!=-1 ) continue ;
    Lucas1(N,P,Q,m,U,V) ;
//cout << endl << "Lucas " << U ;
//V=U ;
    if( U%N!=0 ) continue ;
    Lucas1(N,P,Q,q,U,V) ;
//cout << " " << U << " gcd " << gcd(U,N) ;
//V=U ;
    if( gcd(U,N)==1 ) return true ;
  }
  return false ;
}

bool ECPP(const Mint& N)
{
  int i ;
  int j,k ;
  int n ;
  Mint R ;
  unsigned long d ;
  unsigned long q1 ;
  unsigned long D1;
  unsigned long D ;
  Mint x,y,y2 ;
  Mint wx,wy ;
  int ndiv ;
  Mint m ;
  vector< Mint > mary;
  vector< Mint > dary ;
  Mint t ;
  Fint wf ;
  PolyM T,T1 ;
  Mint g ;
  Mint seed=time(NULL) ;
  Mint J ;
  Mint c ;
  Mint w ;
  Mint q ;
  Mint P1x,P1y ;
  Mint P2x,P2y ;
  vector< Mint > roots ;
  unsigned long h1 ;

  unsigned short* ptbl ;
  unsigned long* rtbl ;
//  unsigned long tblMax=400000 ;
  unsigned long tblMax=  100000 ;
//  unsigned long tblMax=  1000000 ;
  unsigned long Pmax ;
  double suma ;

  if( gcd(N%6,6)>1 ) return false ;

Initilaize:
  j=0 ;
  n=0 ;
  Nj=N ;
#if 1

  ptbl=new unsigned short[tblMax] ;  // prims table
  rtbl=new unsigned long[tblMax] ;   // residue table
  q1=0 ;
  for( k=0 ; k<tblMax ; k++ ) {
    d=q1 ;
    q1=nextprm(q1) ;                 // difference of  primes
    ptbl[k]=q1-d ;
  }
  cout << endl << "max prime " << q1 ;
#endif
Is_Nj_small:
#if 1
  cout << endl << "N " << Nj ;
#endif
#if 1
  q1=0 ;
  for( k=0 ; k<tblMax ; k++ ) {
    q1+=ptbl[k] ;
    rtbl[k]=(Nj+1)%q1 ;            // mod(N+1,p)
  }
#endif

// ŕʂ̔@
#if 0
  for( k=0 ; k<tblMax ; k++ ) {
    q1+=ptbl[k] ;
    if( rtbk[k]=q1-1 ) goto BackTrack ;
  }
#endif

//  d=prmdiv(Nj,10000000) ;
  d=prmdiv(Nj) ;
  if( d==1 ) {
    delete [] ptbl ;
    delete [] rtbl ;
    return true ;
  }
  else if( d!=0 ) {
    goto BackTrack ;
  }

  d=Pocklington(Nj) ;
  if( d==1 ) {
    cout << endl << "Pocklington test " ;
    cout << endl << Nj << " is prime " ;
    delete [] ptbl ;
    delete [] rtbl ;
    return true ;
  }
  m=Nj-1 ;
  divisor2(m,ptbl,rtbl,tblMax,2,dary) ;
  if( dary.size()>=1 ) {
    t=isqrt(m) ;
    q=0 ;
    for( k=0 ; k<dary.size() ; k++ ) {
      if( dary[k]>t && Rabin(dary[k]) ) {
        q=dary[k] ;
        break ;
      }
    }
    if( q!=0 && m%q!=0 ) Error(  m%q!=0 ) ;
    while( !dary.empty() ) dary.pop_back() ;
    // (N-1) test
    if( q!=0 && Pocklington1(Nj,m,q) ) {
      cout << endl << "N-1 test " ;
      cout << endl << "m " << m ;
      cout << endl << "q " << q ;
      cout << endl <<"------------------------" ;
      goto Recurse ;
    }
  }

//cout << endl << "Lucas " ;
  d=Lucas(Nj) ;
  if( d==1 ) {
    delete [] ptbl ;
    delete [] rtbl ;
    return true ;
  }
  m=Nj+1 ;
  divisor2(m,ptbl,rtbl,tblMax,0,dary) ;
  if( dary.size()>=1 ) {
    t=isqrt(m) ;
    q=0 ;
    for( k=0 ; k<dary.size() ; k++ ) {
      if( dary[k]>t && Rabin(dary[k]) ) {
        q=dary[k] ;
        break ;
      }
    }
    if( q!=0 && m%q!=0 ) Error(  m%q!=0 ) ;
    while( !dary.empty() ) dary.pop_back() ;
    // (N+1) test
    if( q!=0 && LucasTest1(Nj,m,q) ) {
      cout << endl << "N+1 test " ;
      cout << endl << "N " << Nj ;
      cout << endl << "m " << m ;
      cout << endl << "q " << q ;
      cout << endl <<"------------------------" ;
      goto Recurse ;
    }
  }

  // Jacobi sum test APR-CL
  if( len10(Nj)<50 && APRCL(Nj) ) {
    cout << endl << "Jacobi sum test " ;
    cout << endl << "Nj " << Nj ;
    cout << endl <<"------------------------" ;
    return true ;
  }

  D=2 ;

cout << endl << "search a good discriminant " ;
  // Ellictic Curve Primality Proving
Choose_next_discriminant: ;

  n++ ;
  for(;;) {
    D=nextDisc(++D) ;
    // solve x^2+D*y^2=4*N
    if( Kronecker(-Mint(D),Nj)!=1 ) continue ;
    if( !Cornacchia4(Mint(D),Nj,x,y) ) continue ;

    // calculate class number
    h1=hDsum1(D,suma) ;
    SetPrecision(suma*M_PI*sqrt(double(D))/9.6+50) ;
    if( h1<=50 ) break ;
  }
cout << " " << D ;
//  cout << endl << "find a solution " ;
//  cout << endl << "x y " << x << " " << y ;
//  cout << " " << (x*x-D*y*y) ;
  if( x*x+D*y*y!=4*Nj ) Error(x*x+D*y*y!=4*Nj) ;


  // enumerate possible group orders 
Factor_m:;
  while( !mary.empty() ) {
    mary.pop_back() ;
  }
  m=Nj+1+x ;
  mary.push_back(m) ;                  // m is possible order
  m=Nj+1-x ;
  mary.push_back(m) ;

  if( D==4 ) {
    m=Nj+1+2*y ;
    mary.push_back(m) ;
    m=Nj+1-2*y ;
    mary.push_back(m) ;
  }
  else if( D==3 ) {
    m=Nj+1+x+3*y ;
    mary.push_back(m) ;
    m=Nj+1-x-3*y ;
    mary.push_back(m) ;
    m=Nj+1+x-3*y ;
    mary.push_back(m) ;
    m=Nj+1-x+3*y ;
    mary.push_back(m) ;
  }



//cout << endl << "push_back" ;

  // decompose possible order
Dose_a_suitable_m_exist: ;
  q=0 ;
  while( q==0 && !mary.empty() ) {
    m=mary.back() ;
    mary.pop_back() ;
    while( !dary.empty() ) dary.pop_back() ;
    wf=Nj ;
    wf=sqrt(sqrt(wf))+1 ;
    t=ceil(wf*wf) ;
#if 0
    ndiv=divisor(m,t,dary) ;
#else
    ndiv=divisor1(m,t,ptbl,rtbl,tblMax,(Nj+1-m),dary) ;
#endif
    for( k=0 ; k<dary.size() ; k++ ) {
      if( dary[k]>t && Rabin(dary[k]) /* && dary[k]<Nj */ ) {
        q=dary[k] ;
        break ;
      }
    }
    while( !dary.empty() ) dary.pop_back() ;
  }
  if( q==0 ) goto Choose_next_discriminant ;
//  else mary.pop_back() ;

#if 0
cout << endl << "m q " <<m << " "<<  q ;
for( i=0 ; i<dary.size() ; i++ )
  cout << " " << dary[i] ;
#endif

  if( m%q!=0 ) Error( m%q!=0 ) ;

Compute_elliptic_curve:
  if( D==4 ) {
    a=-1 ;
    b=0 ;
  }
  else if( D==3 ) {
    a=0 ;
    b=-1 ;
  }
  else {

    // Weber vZ
//    HilbertPolynomial(-D,T1) ;
    Weber(D,T) ;
    while( T[0]==0 ) T>>=1 ;
cout << endl << "Weber Polynomial " << D << " degree=" << deg(T) ;
    for( k=0 ; k<=deg(T) ; k++ ) {
      T[k]%=Nj ;
    }
#if 0
    if( RootModp(T,Nj,roots)==0 ) {
        goto Dose_a_suitable_m_exist ;
    }
#else
    if( D%3!=0 && D%8!=3 ) {
      RootModp(T,Nj,roots) ;
    }
    else if( D%8==3 && D%6!=3 ) {
      RootModp1(T,Nj,roots) ;
    }
    else if( D%6==3 ) RootModp1(T,Nj,roots) ;
    else {
      if( D%3!=0 ) RootModp(T,Nj,roots);
      else RootModp1(T,Nj,roots) ;
    }

    if( roots.size()==0 ) goto Choose_next_discriminant ;
#endif

    if( !roots.empty() ) {
      JmodN(roots[0],Nj,D,J) ;             // j mod p
      while( !roots.empty() ) roots.pop_back() ;
    }

    c=mod(J*invmodp(1728-J,Nj),Nj) ;        // j/(1728-j)
    a=mod(3*c,Nj) ;
    b=mod(2*c,Nj) ;
  }
//  cout << endl << mod(b*invmodp(bb,Nj),Nj) ;
//return true ;
Find_g:
  // search qudratic non-residue elelement in F_N
  for(;;) {
    g=irandom(seed,Nj) ;
    if( Kronecker(g,Nj)!=-1 ) continue ;
    if( D==3 ) {
      if( power(g,(Nj-1)/3,Nj)!=1 ) break ;
    }
    else break ;
  }
  k=0 ;

  // ȉ~Ȑ̓_
Find_P: ;
  do {
    x=irandom(seed,Nj) ;
    y=x*x*x+a*x+b ;
  } while( Kronecker(y,Nj)<0 ) ;
  y=sqrt(y,Nj) ;
  if( (y*y-x*x*x-a*x-b)%Nj!=0 ) Error((y*y-x*x*x-a*x-b)=0) ;
//  k=0 ;
//cout << endl << "P0 (" << x << "," << y << ")" ;

Find_right_curve:

  P2x=x ;
  P2y=y ;
  if( !mult(a,Nj,m/q,P2x,P2y,w) ) {
    cout << endl << "2 goto BackTrack " ;
//Error(order is m) ;
    cout << endl << "1 goto BackTrack " ;
    goto BackTrack  ;
  }
#if 0
cout << endl << "[m/q]P = (" << P2x << "," << P2y << ")" ;
#endif
if( (P2y*P2y-P2x*P2x*P2x-a*P2x-b)%Nj!=0 ) Error(a7a7a) ;
  P1x=P2x ;
  P1y=P2y ;
  if( !mult(a,Nj,q,P1x,P1y,w) ) {
    cout << endl << "elliptic curve test" ;
//    cout << endl << "y^2=x^3+" << a << "*x+" << b ;
    cout << endl << "discriminant " << D ;
    cout << endl << "a = " << a ;
    cout << endl << "b = " << b ;
    cout << endl << "P=(" << x << "," << y << ")" ;
    cout << endl << "m " << m ;
    cout << endl << "q " << q ;
    cout << endl << "[m/q]P = (" << P2x << "," << P2y << ")" ;
    cout << endl << "[m]P = O" ;
    cout << endl <<"------------------------" ;
    if( P1x==0 ) goto Check_P ;
//  cout << endl << "2 goto BackTrack " ;
//Error(order is m) ;
    goto BackTrack  ;
  }

//P1x=x ;
//P1y=y ;
//if( !mult(a,Nj,m,P1x,P1y,w) ) goto BackTrack  ;
//cout << endl << " post *m.q 3*"  ;

//P1x%=Nj ;
#if 0
cout << endl << "[m]P = (" << P1x << "," << P1y << ")" ;
cout << endl <<"------------------------" ;
#endif
//if( P1x%Nj==0 ) cout << endl << "unit " ;

//cout << endl  << P1x ;
  if( P1x%Nj==0 ) goto Check_P ;

  x=P1x ;
  y=P1y ;

Step10:
  k++ ;
  if( wD(D)<=k ) {
    if( !mary.empty() ) goto Dose_a_suitable_m_exist ;
    else goto BackTrack ;
  }
  if( D>4 ) {
    a=mod(a*g*g,Nj) ;
    b=mod(b*g*g*g,Nj) ;
  }
  else if( D==4 ) {
    a=mod(a*g,Nj) ;
  }
  else if( D==3 ) {
    b=mod(b*g,Nj) ;
  }
  goto Find_P ;
//  goto Find_right_curve ;
Find_a_new_P:
    do {
      x=irandom(seed,Nj) ;
      y=((x*x*x)%Nj+a*x+b)%Nj ;
    } while( Kronecker(y,Nj)<0 ) ;

    y=sqrt(y,Nj) ;
    if( x==0 ) {
      delete [] ptbl ;
      delete [] rtbl ;
      return true ;
    }
    goto Step10 ;


Check_P:;
  if( P2x%Nj==0 ) goto Find_a_new_P ;
//  else return true ;
Recurse:
   j++ ;
   Nj=q ;
   goto Is_Nj_small ;
BackTrack: ;
//cout << endl << "BackTrack " ;
  if( j==0 ) {
    delete [] ptbl ;
    delete [] rtbl ;
    return false ;
  }
  else {
    j-- ;
    goto Choose_next_discriminant ;
  }
}




int main(void)
{

//display_rho=true ;
  cout << endl << _Copyright ;
  vector< Fp > root ;
  Mint N="= 2243297958611430246424701735624325623520733109713213853894475007" ;
  N="44407063158353648719572831190575042665156094016138014731570981" ;

//  N=1 ;
//  N<<=160 ;
//  N+=7 ;
//  N=1 ;
//  N<<=100 ;
//  N+=277 ;
//  N="471064017714648581743716115253" ;
//  N="781221660082682887337352611537" ;
  N="44407063158353648719572831190575042665156094016138014731570981" ;
N = "6515894670747361489609591699870445119964416876261985066154614445178435234970298717942724509886480771" ;
//N="1384435372850622112932804334308326689651568940268408537";
//  N="2243297958611430246424701735624325623520733109713213853894475007" ;
//  N="1242679608952272628372591359413759848950066057127148649961499441693920863762185527" ;
//N=" 65670822302684064701133181659108281791806006666541" ;
//N="1881069371691857158934235888672126564551712318446441" ;
//N="28485319769302136014480175348032626270825403782131" ;
//N="37044208456974047045729807761087531209210189733334533" ;
#if 1
  char str[3000] ;
  for(;;) {
    cout << endl << "Input number " << endl ;
    cin >> str ;
    N=parse(str) ;
cout << endl << N ;
    if( N==0 ) break ;
    if( !Rabin(N) ) continue ;
    CPUTime cputime ;
    cputime.Start() ;
    if( ECPP(N) ) cout << endl << N << " is prime ";
    else cout << endl << N << " is not prime ";
    cputime.Stop() ;
    cout << endl << "cputime " << cputime ;
  }
#endif

#if 0

  for( unsigned long n=500 ; n<1000 ; n++ ) {
//    N=(power(Mint(12),n)-1)/11 ;
    N=(power(Mint(11),n)-1)/10;
cout << n << " " ;
    if( !Rabin(N) ) continue ;
//  if( APRCL(N) ) cout << " prime " ;
    CPUTime cputime ;
    cputime.Start() ;
    if( ECPP(N) ) cout << endl << N << " is prime ";
    else cout << endl << N << " is not prime ";
    cputime.Stop() ;
    cout << endl << "cputime " << cputime ;
break ;
  }
#endif
}

