//
//  multilple precision integer class  Ver 2.40
//
//  Copyright (C) by S.Tomabechi 1993,1995
//     All rights reserved.
//
//  constructors and allocating memory
//
#define _ONLY_MINTDECL_

#include "Mint.hpp"
#include <stdlib.h>

#include "Error.hpp"

//
// constructors
//
Mint::Mint(const unsigned long n):digit(NULL),fDigit(0)
{
  allocate(1) ;
  *this=n ;
}

Mint::Mint(const int n):digit(NULL),fDigit(0)
{
  allocate(1) ;
  *this=n ;
}

Mint::Mint():digit(NULL),fDigit(0)      // default constructor
{
  allocate(1) ;
  nDigit=0 ;                     // set 0 as default value 0
}

Mint::Mint(const long n):digit(NULL),fDigit(0)
{
  allocate(1) ;
  operator=(n) ;
}


//  allocate memory to save integer
//
//  return value   true   :  succeeded
//                 false  :  failed
//
//  i[̈m
//  ߂l : true(1) mۂɐꍇ
//         : false(0) mۂɎsꍇ A݂͎sꍇ
//          G[bZ[Wo͂ďI(exit).
//  AmۂɎsꍇAO𑗏o悤ɕύX(?).
//
//  ̊֐ł́Cz̑傫̂݊Ǘ̑̃of|^͊ǗȂ
//
bool Mint::allocate(int size)
{
#ifdef _MEMMAN_DEBUG
int j ;
if( size<=0 ) Error(size<0 ) ;
#endif
  if( size>MAX_DIGITS ) {
    cerr << endl  << "size is " << size << flush  ;
    Error(too big) ;  // requested size is too big
  }

  if( digit ) {             // memry is already allocated

//    if( nDim>size ) return TRUE ;       // enough space

    (nDim=(++size))|=0x01000000 ;

#ifdef _MEMMAN
    _MemMan.free(digit) ;
    digit=(unsigned long*)_MemMan.alloc(NBYTE*size) ;
for( j=0 ; j<size ; j++ ) digit[j]=0xffffffffUL ;   // set magic number for debug
//_MemMan.check(__FILE__,__LINE__) ;
#else
    delete [] digit ;
    digit=new unsigned long[size] ;
#endif
  }
  else {
    (nDim=(++size))|=0x01000000 ;
#ifdef _MEMMAN
    digit=(unsigned long*)_MemMan.alloc(NBYTE*size) ;
for( j=0 ; j<size ; j++ ) digit[j]=0xffffffffUL ;
#else
    digit=new unsigned long[size] ;
#endif
  }

  if( !digit ) {             // failed to allocate
#ifdef _MEMMAN_DEBUG
    _MemMan.debug() ;
#endif
    cerr << "requested size " << size ;
    Error(not enough memory) ;
    return false ;
  }
  else return true ;         // succeed to allocate
}


//
// private constructor
//
Mint::Mint(int size,int fsize):digit(NULL)
{
 if( fsize>0 && Precision>fsize ) {
    size+=(Precision-fsize ) ;
    fsize=Precision ;
  }
  allocate(size) ;
  nDigit=0 ;
  fDigit=fsize ;
}

#ifdef _MEMMAN                 // functions for debugging
void Mint::check(char* file,int line) const
{
#  ifdef _MEMMAN_DEBUG
  if( abs(*((long*)digit-1))!=ndim()+1 ) {
    component(cerr,*this) ;
    cerr << endl << digit ;
    cerr << endl << "file " << file << " line " << line << flush ;
    cerr << endl << "nDigit nDim " << nDigit << " " << nDim
         << "(" << hex << nDim  << dec << ")" ;
    Error(exceed array) ;
  }
                      // check signature
  signed char sign ;
  sign=(signed char)(nDim>>24) ;
  if( sign!=1 && sign!=-1 ) {
    cerr << endl << "file " << file << " line " << line << flush ;
    component(cout,*this) ;
    cerr << endl << "sign " << hex << sign << dec << flush ;
    Error(invalid sign) ;
  }

                             // check # digits
  if( ndim()<nDigit ) {
    cerr << endl << "file " << file << " line " << line << flush ;
//    component(cout,*this) ;
    cerr << endl << "nDim nDigit " << (long)(nDim&_DIGITS_) << " "
         << nDigit ;
    Error(nDigit>nDim) ;
  }
                             // check whether the most significant digit is not zero
  if( nDigit!=0 && digit[nDigit-1]==0 ) {
    component(cerr,*this) ;
    cerr << endl << "file " << file << " line " << line << flush ;
    Error(significant digit=0) ;
  }
#  endif
}
#endif


