//
//  Copyright (C) 1994 by S.Tomabechi.
//  All rights reserved.
//
//  Rabin-Miller primality test
//
//  return value   true  : probably prime
//                 false : not prime
//

#include "Mint.hpp"
#include "MintFunc.hpp"
#include "MBit.hpp"
#include <time.h>

bool Rabin(const Mint& N,unsigned long maxloop)
{
  unsigned long a=time(NULL) ;
  Mint b ;
  Mint q ;
  unsigned long m ;
  long e,t ;
  unsigned long iloop ;
//  int maxloop=5 ;       //  20 ä¿

                          // trial division when N<2^32
  if( N.ndigit()<2 ) return prmdiv(N)==1 ;

                          // even ?
  if( Even(N) )  return N==2 ;

  q=N-1 ;
  t=Val2(q) ;             // 2^t || q   gcd(2,q/(2^t))=1
  if( t!=0 ) q>>=t ;      // divide by 2^t

  for( iloop=maxloop ; (--iloop)>0 ; ) {

    do {
      m=irandom(a) ;
    } while( m<=1 ) ;

    b=apower(Mint(m),q,N) ;  // m^q mod N

    e=0 ;
    if( b==1 ) continue ;    // Fermat's little theorem

    while( b!=1 && b!=-1 && e<=t-2 ) {
      b=square(b)^N ;        // operator^() : the least absolute residue
      e++ ;
    }

    if( b!=-1 ) return false ;

  }
  return true ;
}
